/* $Id: anal.c,v 1.262 2006/11/03 08:23:02 kb Exp $

Copyright (C) 2000  The PARI group.

This file is part of the PARI/GP package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/*******************************************************************/
/*                                                                 */
/*                  SYNTACTICAL ANALYZER FOR GP                    */
/*                                                                 */
/*******************************************************************/
#include "pari.h"
#include "paripriv.h"
#include "anal.h"
#include "parse.h"

static ulong  number(int *pn, char **ps);

static entree *installep(void *f,char *name,long l,long v,long add,entree **table);
static const long SIZEOF_VAR_POLS = 4*sizeof(long);

/* points to the start of the string that remains to be parsed */
static THREAD char *pari_lex_start = NULL;
THREAD char *pari_lasterror, *pari_unusedchar;

/* when non-NULL, points to the entree of a new user function (currently
 * being checked). Used by the compatibility engine in the following way:
 *   when user types in a function whose name has changed, it is understood
 *   as EpNEW; first syntax error (missing = after function definition
 *   usually) triggers err_new_fun() if check_new_fun is set. */
#define NOT_CREATED_YET ((entree *)0x1L)
#define initial_value(ep) ((ep)+1)

/* Mnemonic codes parser:
 *
 * TEMPLATE is assumed to be ";"-separated list of items.  Each item
 * may have one of the following forms: id=value id==value id|value id&~value.
 * Each id consists of alphanum characters, dashes and underscores.
 * IDs are case-sensitive.

 * ARG consists of several IDs separated by punctuation (and optional
 * whitespace).  Each modifies the return value in a "natural" way: an
 * ID from id=value should be the first in the sequence and sets RETVAL to
 * VALUE (and cannot be negated), ID from id|value bit-ORs RETVAL with
 * VALUE (and bit-ANDs RETVAL with ~VALUE if negated), ID from
 * id&~value behaves as if it were noid|value, ID from
 * id==value behaves the same as id=value, but should come alone.

 * For items of the form id|value and id&~value negated forms are
 * allowed: either when arg looks like no[-_]id, or when id looks like
 * this, and arg is not-negated. */

enum { A_ACTION_ASSIGN, A_ACTION_SET, A_ACTION_UNSET };
enum { PARSEMNU_TEMPL_TERM_NL, PARSEMNU_ARG_WHITESP };
#define IS_ID(c)	(isalnum((int)c) || ((c) == '_') || ((c) == '-'))
#define ERR(reason)	STMT_START {	\
    if (failure && first) {		\
	*failure = reason; *failure_arg = NULL; return 0;		\
    } else pari_err(talker,reason); } STMT_END
#define ERR2(reason,s)	STMT_START {	\
    if (failure && first) {		\
	*failure = reason; *failure_arg = s; return 0;		\
    } else pari_err(talker,reason,s); } STMT_END

ulong
parse_option_string(char *arg, char *tmplate, long flag, char **failure, char **failure_arg)
{
    ulong retval = 0;
    char *etmplate = NULL;

    if (flag & PARSEMNU_TEMPL_TERM_NL)
	etmplate = strchr(tmplate, '\n');
    if (!etmplate)
	etmplate = tmplate + strlen(tmplate);

    if (failure)
	*failure = NULL;
    while (1) {
	long numarg;
	char *e, *id;
	char *negated;			/* action found with 'no'-ID */
	int negate;			/* Arg has 'no' prefix removed */
	ulong l, action = 0, first = 1, singleton = 0;
        char *buf, *inibuf;
	static char b[80];

	if (flag & PARSEMNU_ARG_WHITESP)
	    while (isspace((int)*arg)) arg++;
	if (!*arg)
	    break;
	e = arg;
	while (IS_ID(*e)) e++;
	/* Now the ID is whatever is between arg and e. */
	l = e - arg;
	if (l >= sizeof(b))
	    ERR("id too long in a stringified flag");
	if (!l)				/* Garbage after whitespace? */
	    ERR("a stringified flag does not start with an id");
	strncpy(b, arg, l);
	b[l] = 0;
	arg = e;
	e = inibuf = buf = b;
	while (('0' <= *e) && (*e <= '9'))
	    e++;
	if (*e == 0)
	    ERR("numeric id in a stringified flag");	
	negate = 0;
	negated = NULL;
      find:
	id = tmplate;
	while ((id = strstr(id, buf)) && id < etmplate) {
	    if (IS_ID(id[l])) {		/* We do not allow abbreviations yet */
		id += l;		/* False positive */
		continue;
	    }
	    if ((id >= tmplate + 2) && (IS_ID(id[-1]))) {
		char *s = id;

		if ( !negate && s >= tmplate+3
		     && ((id[-1] == '_') || (id[-1] == '-')) )
		    s--;
		/* Check whether we are preceeded by "no" */
		if ( negate		/* buf initially started with "no" */
		     || (s < tmplate+2) || (s[-1] != 'o') || (s[-2] != 'n')
		     || (s >= tmplate+3 && IS_ID(s[-3]))) {
		    id += l;		/* False positive */
		    continue;
		}
		/* Found noID in the template! */
		id += l;
		negated = id;
		continue;		/* Try to find without 'no'. */
	    }
	    /* Found as is */
	    id += l;
	    break;
	}
	if ( !id && !negated && !negate
	     && (l > 2) && buf[0] == 'n' && buf[1] == 'o' ) {
	    /* Try to find the flag without the prefix "no". */
	    buf += 2; l -= 2;
	    if ((buf[0] == '_') || (buf[0] == '-')) { buf++; l--; }
	    negate = 1;
	    if (buf[0])
		goto find;
	}
	if (!id && negated) {	/* Negated and AS_IS forms, prefer AS_IS */
	    id = negated;	/* Otherwise, use negated form */
	    negate = 1;
	}
	if (!id)
	    ERR2("Unrecognized id '%s' in a stringified flag", inibuf);
	if (singleton && !first)
	    ERR("Singleton id non-single in a stringified flag");
	if (id[0] == '=') {
	    if (negate)
		ERR("Cannot negate id=value in a stringified flag");
	    if (!first)
		ERR("Assign action should be first in a stringified flag");
	    action = A_ACTION_ASSIGN;
	    id++;
	    if (id[0] == '=') {
		singleton = 1;
		id++;
	    }
	} else if (id[0] == '^') {
	    if (id[1] != '~')
		pari_err(talker, "Unrecognized action in a template");
	    id += 2;
	    if (negate)
		action = A_ACTION_SET;
	    else
		action = A_ACTION_UNSET;
	} else if (id[0] == '|') {
	    id++;
	    if (negate)
		action = A_ACTION_UNSET;
	    else
		action = A_ACTION_SET;
	}

	e = id;

	while ((*e >= '0' && *e <= '9')) e++;
	while (isspace((int)*e))
	    e++;
	if (*e && (*e != ';') && (*e != ','))
	    pari_err(talker, "Non-numeric argument of an action in a template");
	numarg = atol(id);		/* Now it is safe to get it... */
	switch (action) {
	case A_ACTION_SET:
	    retval |= numarg;
	    break;
	case A_ACTION_UNSET:
	    retval &= ~numarg;
	    break;
	case A_ACTION_ASSIGN:
	    retval = numarg;
	    break;
	default:
	    ERR("error in parse_option_string");
	}
	first = 0;
	if (flag & PARSEMNU_ARG_WHITESP)
	    while (isspace((int)*arg))
		arg++;
	if (*arg && !(ispunct((int)*arg) && *arg != '-'))
	    ERR("Junk after an id in a stringified flag");
	/* Skip punctuation */
	if (*arg)
	    arg++;
    }
    return retval;
}

char*
get_origin(void) { return pari_lex_start; }

static void
unused_chars(char *analyseur, char *c, int strict)
{
  long n = 2 * term_width() - (17+19+1); /* Warning + unused... + . */
  char *s;
  if (strict) pari_err(talker2,"unused characters", analyseur, c);
  if ((long)strlen(analyseur) > n)
  {
    s = gpmalloc(n + 1);
    n -= 5;
    (void)strncpy(s,analyseur, n);
    strcpy(s + n, "[+++]");
  }
  else
    s = pari_strdup(analyseur);
  pari_warn(warner, "unused characters: %s", s);
  gpfree(s);
}

INLINE void
seq_init(char *lex, int strict)
{
  pari_lex_start = lex;
  reset_break();
  pari_unusedchar=NULL;
  if (pari_parse(&lex))
  {
    if (pari_unusedchar)
      unused_chars(pari_unusedchar,pari_lex_start,strict);
    else  
      pari_err(talker2,pari_lasterror,lex-1,pari_lex_start);
  }
}

#define HANDLE_FOREIGN(t)\
  if (foreignExprHandler && *t == foreignExprSwitch)\
    return (*foreignExprHandler)(t);

GEN
readseq(char *t)
{
  pari_sp av = top - avma;
  GEN z;

  HANDLE_FOREIGN(t);

  seq_init(t,0); z = evalnode();
  av = top - av; /* safer than recording av = avma: f() may call allocatemem */
  if (isclone(z)) { avma = av; return gcopy(z); }
  return gerepileupto(av, z);
}

/* filtered readseq = remove blanks and comments */
GEN
gp_read_str(char *s)
{
  char *t = filtre(s, (compatible == OLDALL));
  GEN x = readseq(t);
  gpfree(t); return x;
}

/* check syntax, then execute */
GEN
gpreadseq(char *t, int strict)
{
  seq_init(t, strict); 
  return evalnode();
}

long
check_proto(char *code)
{
  long arity = 0;
  char *s = code, *old;
  if (*s == 'l' || *s == 'v' || *s == 'i') s++;
  while (*s && *s != '\n') switch (*s++)
  {
    case '&':
    case 'G':
    case 'I': case 'E':
    case 'L':
    case 'M':
    case 'P':
    case 'S':
    case 'V':
    case 'f':
    case 'n':
    case 'p':
    case 'r':
    case 'x': 
      arity++;
      break;
    case 's':
      if (*s == '*') s++;
      arity++;
      break;
    case 'D':
      if (*s == 'G' || *s == '&' || *s == 'n' || *s == 'I' || *s == 'V')
      {
        arity++;
        s++; break;
      }
      old = s; while (*s != ',') s++;
      if (*s != ',') pari_err(talker2, "missing comma", old, code);
      break;
    case '=':
    case ',': break;
    case '\n': return arity; /* Before the mnemonic */

    case 'l':
    case 'i':
    case 'v': pari_err(talker2, "this code has to come first", s-1, code);
    default: pari_err(talker2, "unknown parser code", s-1, code);
  }
  return arity;
}

entree *
install(void *f, char *name, char *code)
{
  long hash, arity;
  entree *ep = is_entry_intern(name, functions_hash, &hash);

  arity=check_proto(code);
  if (ep)
  {
    if (ep->valence != EpINSTALL)
      pari_err(talker,"[install] identifier '%s' already in use", name);
    pari_warn(warner, "[install] updating '%s' prototype; module not reloaded", name);
    if (ep->code) gpfree(ep->code);
  }
  else
  {
    char *s = name;
    if (isalpha((int)*s))
      while (is_keyword_char(*++s)) /* empty */;
    if (*s) pari_err(talker2,"not a valid identifier", s, name);
    ep = installep(f, name, strlen(name), EpINSTALL, 0, functions_hash + hash);
  }
  ep->code = pari_strdup(code);
  ep->arity=arity;
  return ep;
}

/*******************************************************************/
/*                                                                 */
/*                            VARIABLES                            */
/*                                                                 */
/*******************************************************************/
/* As a rule, ep->value is a clone (COPY). push_val and pop_val are private
 * functions for use in sumiter: we want a temporary ep->value, which is NOT
 * a clone (PUSH), to avoid unnecessary copies. */

/* ep->args is the stack of old values (INITIAL if initial value, from
 * installep) */
typedef struct var_cell {
  struct var_cell *prev; /* cell associated to previous value on stack */
  GEN value; /* last value (not including current one, in ep->value) */
  char flag; /* status of _current_ ep->value: PUSH or COPY ? */
} var_cell;
#define INITIAL NULL
/* Push x on value stack associated to ep. Assume EpVALENCE(ep)=EpVAR/EpGVAR */
void
new_val_cell(entree *ep, GEN x, char flag)
{
  var_cell *v = (var_cell*) gpmalloc(sizeof(var_cell));
  v->value  = (GEN)ep->value;
  v->prev   = (var_cell*) ep->pvalue;
  v->flag   = flag;

  /* beware: f(p) = Nv = 0
   *         Nv = p; f(Nv) --> this call would destroy p [ isclone ] */
  ep->value = (flag == COPY_VAL)? gclone(x):
                                  (x && isclone(x))? gcopy(x): x;
  /* Do this last. In case the clone is <C-C>'ed before completion ! */
  ep->pvalue= (char*)v;
}

void
freeep(entree *ep)
{
  if (foreignFuncFree && ep->code && (*ep->code == 'x'))
    (*foreignFuncFree)(ep); /* function created by foreign interpreter */

  if (EpSTATIC(ep)) return; /* gp function loaded at init time */
  if (ep->help) gpfree(ep->help);
  if (ep->code) gpfree(ep->code);
  switch(EpVALENCE(ep))
  {
    case EpVAR:
    case EpGVAR:
      while (ep->pvalue) pop_val(ep);
      break;
    case EpUSER:
      break;
    case EpALIAS:
      gunclone((GEN)ep->value); break;
  }
  gpfree(ep);
}

void
push_val(entree *ep, GEN a) { new_val_cell(ep, a, PUSH_VAL); }

/* kill ep->value and replace by preceding one, poped from value stack */
void
pop_val(entree *ep)
{
  var_cell *v = (var_cell*) ep->pvalue;

  if (v == INITIAL) return;
  if (v->flag == COPY_VAL) killbloc((GEN)ep->value);
  ep->value  = v->value;
  ep->pvalue = (char*) v->prev;
  gpfree((void*)v);
}

/* as above IF ep->value was PUSHed, or was created after block number 'loc'
   return 0 if not deleted, 1 otherwise [for recover()] */
int
pop_val_if_newer(entree *ep, long loc)
{
  var_cell *v = (var_cell*) ep->pvalue;

  if (v == INITIAL) return 0;
  if (v->flag == COPY_VAL && !pop_entree_bloc(ep, loc)) return 0;
  ep->value = v->value;
  ep->pvalue= (char*) v->prev;
  gpfree((void*)v); return 1;
}

/* set new value of ep directly to val (COPY), do not save last value unless
 * it's INITIAL. */
void
changevalue(entree *ep, GEN x)
{
  var_cell *v = (var_cell*) ep->pvalue;
  if (v == INITIAL) new_val_cell(ep, x, COPY_VAL);
  else
  {
    x = gclone(x); /* beware: killbloc may destroy old x */
    if (v->flag == COPY_VAL) killbloc((GEN)ep->value); else v->flag = COPY_VAL;
    ep->value = (void*)x;
  }
}

/* as above, but PUSH, not COPY */
void
changevalue_p(entree *ep, GEN x)
{
  var_cell *v = (var_cell*) ep->pvalue;
  if (v == INITIAL) new_val_cell(ep,x, PUSH_VAL);
  else
  {
    if (v->flag == COPY_VAL) { killbloc((GEN)ep->value); v->flag = PUSH_VAL; }
    ep->value = (void*)x;
  }
}

/* make GP variables safe for avma = top */
void
var_make_safe(void)
{
  long n;
  entree *ep;
  for (n = 0; n < functions_tblsz; n++)
    for (ep = functions_hash[n]; ep; ep = ep->next)
      if (EpVALENCE(ep) == EpVAR)
      { /* make sure ep->value is a COPY */
        var_cell *v = (var_cell*)ep->pvalue;
        if (v && v->flag == PUSH_VAL) {
          GEN x = (GEN)ep->value;
          if (x) changevalue(ep, (GEN)ep->value); else pop_val(ep);
        }
      }
}

/* n = hashvalue(ep->name), delete ep */
void
kill_from_hashlist(entree *ep, long n)
{
  entree *e = functions_hash[n];
  if (e == ep) { functions_hash[n] = ep->next; return; }
  for (; e; e = e->next)
    if (e->next == ep) { e->next = ep->next; return; }
}

#define ALIAS(ep) (entree *) ((GEN)ep->value)[1]

/* kill aliases pointing to EP */
static void 
kill_alias(entree *EP)
{
  entree *ep, *epnext;
  long n;
  for (n = 0; n < functions_tblsz; n++)
    for (ep = functions_hash[n]; ep; ep = epnext)
    {
      epnext = ep->next;
      if (EpVALENCE(ep) == EpALIAS && EP == ALIAS(ep)) kill0(ep);
    }
}

/* Kill entree ep, i.e free all memory it occupies, remove it from hashtable.
 * If it's a variable set a "black hole" in varentries[v]. x = 0-th variable
 * can NOT be killed (only the value) */
void
kill0(entree *ep)
{
  const char *s = ep->name;
  long v;

  if (EpSTATIC(ep))
    pari_err(talker,"can't kill that");
  switch(EpVALENCE(ep))
  {
    case EpVAR:
    case EpGVAR:
      while (ep->pvalue) pop_val(ep);
      return;
      v = varn(ep->value); if (!v) return; /* never kill x */
      varentries[v] = NULL; break;
    case EpUSER: kill_alias(ep);
      break;
  }
  kill_from_hashlist(ep, hashvalue(s));
  freeep(ep);
}

void
addhelp(entree *ep, char *s)
{
  if (ep->help && !EpSTATIC(ep)) gpfree(ep->help);
  ep->help = pari_strdup(s);
}

GEN
type0(GEN x)
{
  const char *s = type_name(typ(x));
  return strtoGENstr(s);
}

/*******************************************************************/
/*                                                                 */
/*                              PARSER                             */
/*                                                                 */
/*******************************************************************/

static ulong
number(int *n, char **s)
{
  ulong m = 0;
  for (*n = 0; *n < 9 && isdigit((int)**s); (*n)++,(*s)++)
    m = 10*m + (**s - '0');
  return m;
}

ulong
u_pow10(int n)
{
  const ulong pw10[] = { 1UL, 10UL, 100UL, 1000UL, 10000UL, 100000UL,
                        1000000UL, 10000000UL, 100000000UL, 1000000000UL };
  return pw10[n];
}

static GEN
int_read_more(GEN y, char **ps)
{
  pari_sp av = avma;
  int i = 0, nb;
  while (isdigit((int)**ps))
  {
    ulong m = number(&nb, ps);
    if (avma != av && ++i > 4) { avma = av; i = 0; } /* HACK gerepile */
    y = addumului(m, u_pow10(nb), y);
  }
  return y;
}

static long
exponent(char **pts)
{
  char *s = *pts;
  long n;
  int nb;
  switch(*++s)
  {
    case '-': s++; n = -(long)number(&nb, &s); break;
    case '+': s++; /* Fall through */
    default: n = (long)number(&nb, &s);
  }
  *pts = s; return n;
}

static GEN
real_0_digits(long n) {
  long b = (n > 0)? (long)(n/L2SL10): (long)-((-n)/L2SL10 + 1);
  return real_0_bit(b);
}

static GEN
real_read(pari_sp av, char **s, GEN y, long PREC)
{
  long l, n = 0;
  switch(**s)
  {
    default: return y; /* integer */
    case '.':
    {
      char *old = ++*s;
      if (isalpha((int)**s))
      {
        if (**s == 'E' || **s == 'e') {
          n = exponent(s);
          if (!signe(y)) { avma = av; return real_0_digits(n); }
          break;
        }
        --*s; return y; /* member */
      }
      y = int_read_more(y, s);
      n = old - *s;
      if (**s != 'E' && **s != 'e')
      {
        if (!signe(y)) { avma = av; return real_0(PREC); }
        break;
      }
    }
    /* Fall through */
    case 'E': case 'e':
      n += exponent(s);
      if (!signe(y)) { avma = av; return real_0_digits(n); }
  }
  l = lgefint(y); if (l < (long)PREC) l = (long)PREC;
  if (!n) return itor(y, l);
  y = itor(y, l+1);
  if (n > 0)
    y = mulrr(y, rpowuu(10UL, (ulong)n, l+1));
  else
    y = divrr(y, rpowuu(10UL, (ulong)-n, l+1));
  return gerepileuptoleaf(av, rtor(y, l));
}

static GEN
int_read(char **s)
{
  int nb;
  GEN y = utoi(number(&nb, s));
  if (nb == 9) y = int_read_more(y, s);
  return y;
}

GEN
strtoi(char *s) { return int_read(&s); }

GEN 
strtor(char *s, long prec)
{
  pari_sp av = avma;
  GEN y = int_read(&s);
  y = real_read(av, &s, y, prec);
  if (typ(y) == t_REAL) return y;
  return gerepileuptoleaf(av, itor(y, prec));
}

static void
skipdigits(char **lex) {
  while (isdigit((int)**lex)) ++*lex;
}

static int
skipexponent(char **lex) {
  if (**lex=='e' || **lex=='E')
  {
    ++*lex;
    if ( **lex=='+' || **lex=='-' ) ++*lex;
    skipdigits(lex);
    return 1;
  }
  return 0;
}

static int
skipconstante(char **lex)
{
  skipdigits(lex);
  if (**lex=='.')
  {
    char *old = ++*lex;
    if (isalpha((int)**lex))
    {
      skipexponent(lex);
      if (*lex == old)
      {
        --*lex; /* member */
        return 0;
      }
    }
    skipdigits(lex);
    return 1;
  }
  return skipexponent(lex);
}

static GEN
constante(char **lex)
{
  char *old=*lex;
  if (!skipconstante(lex))
  {
    *lex=old;
    return int_read(lex);
  }
  return strntoGENstr(old, *lex-old);
}

/* inline is_keyword_char(). Not worth a static array. */
#define is_key(c) (isalnum((int)(c)) || (c)=='_')

long
is_keyword_char(char c) { return is_key(c); }

static void
skipstring(char **lex)
{
  while (**lex)
  {
    while (**lex == '\\') *lex+=2;
    if (**lex == '"')
    {
      if ((*lex)[1] != '"') break;
      *lex += 2; continue;
    }
    (*lex)++;
  }
}

int pari_lex(union token_value *yylval, struct node_loc *yylloc, char **lex)
{
  yylloc->start=*lex;
  if (!**lex) 
  {
    yylloc->end=*lex;   
    return 0;
  }
  if (isalpha((int)**lex)) 
  {
    while (is_key(**lex))
      ++*lex;
    yylloc->end=*lex;   
    return KENTRY;
  }
  if (**lex=='"')
  {
    ++*lex;
    skipstring(lex);
    if (!**lex)
      pari_err(talker2,"run-away string",*lex-1,pari_lex_start);
    ++*lex;
    yylloc->end=*lex;   
    return KSTRING;
  }
  if (**lex == '.' &&  !isdigit((int)(*lex)[1]) 
      && (*lex)[1]!='e' && (*lex)[1]!='E')
  {
    ++*lex;
    yylloc->end=*lex;
    return **lex;
  }
  if (isdigit((int)**lex) || **lex == '.') 
  {
    yylval->gen = constante(lex);
    yylloc->end = *lex;
    switch(typ(yylval->gen))
    {
    case  t_INT:
      return KINTEGER;
    case  t_STR:
      return KREAL;
    }
  }
  if ((*lex)[1]=='=')
    switch (**lex)
    {
    case '=':
      *lex+=2; yylloc->end = *lex; return KEQ;
    case '!':
      *lex+=2; yylloc->end = *lex; return KNE;
    case '>':
      *lex+=2; yylloc->end = *lex; return KGE;
    case '<':
      *lex+=2; yylloc->end = *lex; return KLE;
    case '*':
      *lex+=2; yylloc->end = *lex; return KME;
    case '/':
      *lex+=2; yylloc->end = *lex; return KDE;
    case '\\':
      *lex+=2; yylloc->end = *lex; return KEUCE;
    case '%':
      *lex+=2; yylloc->end = *lex; return KMODE;
    case '+':
      *lex+=2; yylloc->end = *lex; return KPE;
    case '-':
      *lex+=2; yylloc->end = *lex; return KSE;
    }
  if (**lex=='\'' && (*lex)[1]=='(')
  {
    *lex+=2; yylloc->end = *lex; return KDER;
  }
  if (**lex=='\\' && (*lex)[1]=='/')
    switch((*lex)[2])
    {
    case '=':
      *lex+=3; yylloc->end = *lex; return KDRE;
    default:
      *lex+=2; yylloc->end = *lex; return KDR;
    }
  if ((*lex)[1]==**lex)
    switch (**lex)
    {
    case '&':
      *lex+=2; yylloc->end = *lex; return KAND;
    case '|':
      *lex+=2; yylloc->end = *lex; return KOR;
    case '+':
      *lex+=2; yylloc->end = *lex; return KPP;
    case '-':
      *lex+=2; yylloc->end = *lex; return KSS;
    case '>':
      if ((*lex)[2]=='=') { *lex+=3; yylloc->end = *lex; return KSRE;}
      *lex+=2; yylloc->end = *lex; return KSR;
    case '<':
      if ((*lex)[2]=='=') 
         { *lex+=3; yylloc->end = *lex; return KSLE; }
      *lex+=2; yylloc->end = *lex; return KSL;
    }
  yylloc->end = *lex+1;
  return *(*lex)++;
}

/********************************************************************/
/**                                                                **/
/**                            STRINGS                             **/
/**                                                                **/
/********************************************************************/

static char *
translate(char **src, char *s)
{
  char *t = *src;
  while (*t)
  {
    while (*t == '\\')
    {
      switch(*++t)
      {
	case 'e':  *s='\033'; break; /* escape */
	case 'n':  *s='\n'; break;
	case 't':  *s='\t'; break;
	default:   *s=*t; if (!*t) pari_err(talker,"unfinished string");
      }
      t++; s++;
    }
    if (*t == '"')
    {
      if (t[1] != '"') break;
      t += 2; continue;
    }
    *s++ = *t++;
  }
  *s=0; *src=t; return s;
}

static void
err_match(char *s, char c)
{
  char str[64];
  sprintf(str,"expected character: '%c' instead of",c);
  pari_err(talker2,str,s,pari_lex_start);
}

#define match2(s,c) if (*s != c) err_match(s,c);

/*  Read a "string" from src. Format then copy it, starting at s. Return
 *  pointer to char following the end of the input string */
char *
readstring(char *src, char *s)
{
  match2(src, '"'); src++; s = translate(&src, s);
  match2(src, '"'); return src+1;
}

/* return the first n0 chars of s as a GEN [s may not be 0-terminated] */
GEN
strntoGENstr(const char *s, long n0)
{
  long n = nchar2nlong(n0+1);
  GEN x = cgetg(n+1, t_STR);
  char *t = GSTR(x);
  strncpy(t, s, n0); t[n0] = 0; return x;
}

GEN
strtoGENstr(const char *s) { return strntoGENstr(s, strlen(s)); }
/********************************************************************/
/**                                                                **/
/**                          READ FUNCTIONS                        **/
/**                                                                **/
/********************************************************************/

entree *
do_alias(entree *ep)
{
  while (ep->valence == EpALIAS) ep = ALIAS(ep);
  return ep;
}

/********************************************************************/
/**                                                                **/
/**                   HASH TABLE MANIPULATIONS                     **/
/**                                                                **/
/********************************************************************/

/* return hashing value for identifier s */
long
hashvalue(const char *s)
{
  long n = 0;
  while (*s) { n = (n<<1) ^ *s; s++; }
  if (n < 0) n = -n;
  return n % functions_tblsz;
}

/* Looking for entry in hashtable. ep1 is the cell's first element */
static entree *
findentry(const char *name, entree *ep1)
{
  entree *ep;

  for (ep = ep1; ep; ep = ep->next)
    if (!strcmp(ep->name, name)) return ep;

  if (foreignAutoload) /* Try to autoload. */
    return foreignAutoload(name);
  return NULL; /* not found */
}

entree *
is_entry(const char *s)
{
  return is_entry_intern(s,functions_hash,NULL);
}

entree *
is_entry_intern(const char *s, entree **table, long *pthash)
{
  long hash = hashvalue(s); 
  if (pthash) *pthash = hash;
  return findentry(s, table[hash]);
}

int
is_identifier(char *s)
{
  while (*s && is_keyword_char(*s)) s++;
  return *s? 0: 1;
}

static entree *
installep(void *f, char *name, long len, long valence, long add, entree **table)
{
  entree *ep = (entree *) gpmalloc(sizeof(entree) + add + len+1);
  const entree *ep1 = initial_value(ep);
  char *u = (char *) ep1 + add;

  ep->name    = u; strncpy(u, name,len); u[len]=0;
  ep->valence = valence;
  ep->value   = f? f: (void *) ep1;
  ep->menu    = 0;
  ep->code    = NULL;
  ep->help    = NULL;
  ep->pvalue  = INITIAL; /* necessary, see var_cell definition */
  ep->arity   = 0;
  ep->lvars   = NULL;
  ep->next    = *table;
  return *table = ep;
}

long
manage_var(long n, entree *ep)
{
  static long max_avail = MAXVARN; /* max variable not yet used */
  static long nvar; /* first GP free variable */
  long var;

  switch(n) {
      case manage_var_init: return nvar=0;
      case manage_var_next: return nvar;
      case manage_var_max_avail: return max_avail;
      case manage_var_pop:
        var = (long)ep;
        if (var != nvar-1) pari_err(talker,"can't pop gp variable");
        return --nvar;
      case manage_var_delete:
	/* user wants to delete one of his/her/its variables */
	if (max_avail == MAXVARN-1) return 0; /* nothing to delete */
	max_avail++;
	return max_avail+1;
      case manage_var_create: break;
      default: pari_err(talker, "panic");
  }

  if (nvar == max_avail) pari_err(talker,"no more variables available");
  if (ep)
  {
    GEN p = (GEN)ep->value;
    var = nvar++;
    /* create pol_x[var] */
    p[0] = evaltyp(t_POL) | evallg(4);
    p[1] = evalsigne(1) | evalvarn(var);
    gel(p,2) = gen_0;
    gel(p,3) = gen_1;
    varentries[var] = ep;
  }
  else
    var = max_avail--;
  return var;
}

long
fetch_var(void)
{
  return manage_var(manage_var_create,NULL);
}

entree *
fetch_named_var(char *s)
{
  entree **funhash = functions_hash + hashvalue(s);
  entree *ep = findentry(s, *funhash);
  if (ep)
  {
    switch (EpVALENCE(ep))
    {
      case EpVAR: case EpGVAR: break;
      default: pari_err(talker, "%s already exists with incompatible valence", s);
    }
    return ep;
  }
  ep = installep(NULL,s,strlen(s),EpVAR, SIZEOF_VAR_POLS, funhash);
  (void)manage_var(manage_var_create,ep); return ep;
}

long
fetch_user_var(char *s)
{
  return varn( initial_value(fetch_named_var(s)) );
}

void
delete_named_var(entree *ep)
{
  (void)manage_var(manage_var_pop, (entree*)varn(initial_value(ep)));
  kill0(ep);
}

long
delete_var(void)
{
  return manage_var(manage_var_delete,NULL);
}

void
name_var(long n, char *s)
{
  entree *ep;
  char *u;

  if (n < manage_var(manage_var_next,NULL))
    pari_err(talker, "renaming a GP variable is forbidden");
  if (n > (long)MAXVARN)
    pari_err(talker, "variable number too big");

  ep = (entree*)gpmalloc(sizeof(entree) + strlen(s) + 1);
  u = (char *)initial_value(ep);
  ep->valence = EpVAR;
  ep->name = u; strcpy(u,s);
  ep->value = gen_0; /* in case geval is called */
  if (varentries[n]) gpfree(varentries[n]);
  varentries[n] = ep;
}

GEN
gpolvar(GEN x)
{
  long v;
  if (!x) {
    long k = 1, n = manage_var(manage_var_next,NULL);
    GEN z = cgetg(n+1, t_VEC);
    for (v = 0; v < n; v++)
    {
      entree *ep = varentries[v];
      if (ep) gel(z,k++) = (GEN)initial_value(ep);
    }
    if (k <= n) {
      setlg(z,k);
      stackdummy((pari_sp)(z+n), (pari_sp)(z+k));
    }
    return z;
  }
  if (typ(x)==t_PADIC) return gcopy( gel(x,2) );
  v = gvar(x);
  if (v==BIGINT) pari_err(typeer,"gpolvar");
  return pol_x(v);
}

/********************************************************************/
/**                                                                **/
/**                        SIMPLE GP FUNCTIONS                     **/
/**                                                                **/
/********************************************************************/

void
alias0(char *s, char *old)
{
  entree *ep, *e;
  long hash;
  GEN x;

  ep = is_entry(old);
  if (!ep) pari_err(talker,"unknown function");
  switch(EpVALENCE(ep))
  {
    case EpVAR: case EpGVAR:
      pari_err(talker,"only functions can be aliased");
  }

  if ( (e = is_entry_intern(s, functions_hash, &hash)) )
  {
    if (EpVALENCE(e) != EpALIAS)
      pari_err(talker,"can't replace an existing symbol by an alias");
    kill0(e);
  }
  ep = do_alias(ep); x = newbloc(2);
  x[0] = evaltyp(t_STR)|evallg(2); /* for getheap */
  gel(x,1) = (GEN)ep;
  (void)installep(x, s, strlen(s), EpALIAS, 0, functions_hash + hash);
}

GEN 
ifpari(GEN g, GEN a/*closure*/, GEN b/*closure*/)
{
  if (gcmp0(g)) /* false */
    return b?closure_eval(b):gnil;
  else /* true */
    return a?closure_eval(a):gnil;
}

GEN
andpari(GEN a, GEN b/*closure*/)
{
  GEN g;
  if (gcmp0(a))
    return gen_0;
  g=closure_eval(b);
  if (!g) return g;
  return gcmp0(g)?gen_0:gen_1;
}

GEN
orpari(GEN a, GEN b/*closure*/)
{
  GEN g;
  if (!gcmp0(a))
    return gen_1;
  g=closure_eval(b);
  if (!g) return g;
  return gcmp0(g)?gen_0:gen_1;
}

GEN gstore(GEN *x, GEN y)
{
  *x=gcopy(y);
  return *x;
}

GEN gmule(GEN *x, GEN y)
{
  *x=gmul(*x,y);
  return *x;
}

GEN gdive(GEN *x, GEN y)
{
  *x=gdiv(*x,y);
  return *x;
}

GEN gdivente(GEN *x, GEN y)
{
  *x=gdivent(*x,y);
  return *x;
}

GEN gdivrounde(GEN *x, GEN y)
{
  *x=gdivround(*x,y);
  return *x;
}

GEN gmode(GEN *x, GEN y)
{
  *x=gmod(*x,y);
  return *x;
}

GEN gshiftle(GEN *x, long n)
{
  *x=gshift(*x,n);
  return *x;
}

GEN gshiftre(GEN *x, long n)
{
  *x=gshift(*x,-n);
  return *x;
}

GEN gadde(GEN *x, GEN y)
{
  *x=gadd(*x,y);
  return *x;
}

GEN gadd1e(GEN *x)
{
  *x=typ(*x)==t_INT?addis(*x,1):gaddgs(*x,1);
  return *x;
}

GEN gsube(GEN *x, GEN y)
{
  *x=gsub(*x,y);
  return *x;
}

GEN gsub1e(GEN *x)
{
  *x=typ(*x)==t_INT?subis(*x,1):gsubgs(*x,1);
  return *x;
}

GEN gshift_right(GEN x, long n) { return gshift(x,-n); }

/********************************************************************/
/**                                                                **/
/**            PRINT USER FUNCTION AND MEMBER FUNCTION             **/
/**                                                                **/
/********************************************************************/

void
print_all_user_fun(void)
{
  entree *ep;
  int i;
  for (i = 0; i < functions_tblsz; i++)
    for (ep = functions_hash[i]; ep; ep = ep->next)
      if (EpVALENCE(ep) == EpUSER)
      {
        pariputc('{'); pariputs(ep->code);
        pariputc('}'); pariputs("\n\n");
      }
}
