/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "anal.h"
#include "tree.h"
#include "opcode.h"
#include "stack.h"

static gp2c_stack s_node, s_value;
node *tree;
value_t *value;

static gp2c_stack s_opcode, s_data;
static long *opcode;
static GEN *data;
static long offset=-1;

const char *opname[]={"_||_", "_&&_", "_==_", "_!=_", "_>=_", "_>_", "_<=_", "_<_", "_-_","_+_","_<<_", "_>>_", "_%_", "_\\/_", "_\\_", "_/_", "_*_","_^_","__","_=_","_--","_++","_-=_", "_+=_", "_<<=_", "_>>=_", "_%=_", "_\\/=_", "_\\=_", "_/=_", "_*=_","+_","-_","!_","_!","_'","_~","%","#_","",""};

void
init_compiler(void)
{
  pari_sp ltop=avma;
  long i;
  stack_init(&s_node,sizeof(*tree),(void **)&tree);
  stack_init(&s_value,sizeof(*value),(void **)&value);
  stack_alloc(&s_value,OPnboperator);
  s_value.n=OPnboperator;
  for (i=0;i<OPnboperator;i++)
  {
    value[i].type=CSTstr;
    value[i].gen = cgetalloc(t_STR, nchar2nlong(strlen(opname[i])+1)+1);
    strcpy(GSTR(value[i].gen), opname[i]);  
  }
  stack_init(&s_opcode,sizeof(*opcode),(void **)&opcode);
  stack_init(&s_data,sizeof(*data),(void **)&data);
  avma=ltop;
}

static void
freenode(void)
{
  s_node.n=0;
  s_value.n=OPnboperator;
}

GEN
evalnode(void)
{
  GEN code, res;
  code=gp_closure(s_node.n-1);
  freenode();
  res=closure_eval(code);
  if (!res) return gnil;
  return res;
}

long newnode(Ffunc f, long x, long y, struct node_loc *loc)
{
  long n=stack_new(&s_node);
  tree[n].f=f;
  tree[n].x=x;
  tree[n].y=y;
  tree[n].start=loc->start;
  tree[n].end=loc->end;
  return n;
}

long newopcall(OPerator op, long x, long y, struct node_loc *loc)
{
  if (y==-1)
    return newnode(Ffunction,op,x,loc);
  else
    return newnode(Ffunction,op,newnode(Flistarg,x,y,loc),loc);
}

static GEN
strntoGENexp(char *start, char *end)
{
  GEN z = cgetg(1+nchar2nlong(end-start), t_STR);
  char *s=GSTR(z),*t=start+1;
  while (t<=end)
  {
    while (*t == '\\')
    {
      switch(*++t)
      {
	case 'e':  *s='\033'; break; /* escape */
	case 'n':  *s='\n'; break;
	case 't':  *s='\t'; break;
	default:   *s=*t; if (!*t) pari_err(talker,"unfinished string");
      }
      t++; s++;
    }
    if (*t == '"')
    {
      if (t[1] != '"') break;
      t += 2; continue;
    }
    *s++ = *t++;
  }
  *s++=0;
  return z;
}

long newstringvalue(struct node_loc *loc)  
{
  long n=stack_new(&s_value);
  value[n].type=CSTstr;
  value[n].gen=strntoGENexp(loc->start,loc->end);
  return n;
}

long newquotevalue(struct node_loc *loc)
{
  long n=stack_new(&s_value);
  value[n].type=CSTquote;
  value[n].gen=strntoGENstr(loc->start,loc->end-loc->start);
  return n;
}

static long newrealvalue(GEN g)
{
  long n=stack_new(&s_value);
  value[n].type=CSTreal;
  value[n].gen=g;
  return n;
}

long newgenvalue(GEN g)
{
  long n=stack_new(&s_value);
  value[n].type=CSTgen;
  value[n].gen=g;
  return n;
}

long newgennode(GEN g, struct node_loc *loc)
{
  if (typ(g)==t_INT)
  {
    long s;
    if (signe(g)==0) return newnode(Fsmall,0,-1,loc);
    if ((s=itos_or_0(g))) return newnode(Fsmall,s,-1,loc);
    return newnode(Fconst,newgenvalue(g),-1,loc);
  }
  return newnode(Fconst,newrealvalue(g),-1,loc);
}

long newentry(struct node_loc *loc)
{
  long n=stack_new(&s_value);
  value[n].type=CSTstr;
  value[n].gen=strntoGENstr(loc->start,loc->end-loc->start);
  return n;
}

long newmember(struct node_loc *loc)
{
  long n=stack_new(&s_value);
  long l=loc->end-loc->start;
  char *t;
  value[n].type= CSTstr;
  value[n].gen = cgetg(nchar2nlong(l+3)+1, t_STR);
  t = GSTR(value[n].gen);
  *t++='_'; *t++='.';
  strncpy(t, loc->start, l); t[l] = 0; 
  return n;
}

void
getcodepos(long *sop, long *sdat, long *off)
{
  *sop=s_opcode.n;
  *sdat=s_data.n;
  *off=offset;
  offset=s_data.n-1;
}

GEN
getclosure(long sop, long sdat, long off)
{
  long lop =s_opcode.n+1-sop;
  long ldat=s_data.n+1-sdat;
  GEN cl=cgetg(3,t_VEC);
  long i;
  gel(cl,1) = cgetg(lop,  t_VECSMALL);
  gel(cl,2) = cgetg(ldat, t_VEC);
  for(i=1;i<lop;i++)
    mael(cl, 1, i) = opcode[i+sop-1];
  s_opcode.n=sop;
  for(i=1;i<ldat;i++)
  {
    gmael(cl, 2, i) = gcopy(data[i+sdat-1]);
    gunclone(data[i+sdat-1]);
  }
  s_data.n=sdat;
  offset=off;
  return cl;
}

void
op_push(op_code o, long x)
{
  long n=stack_new(&s_opcode);
  long m=stack_new(&s_opcode);
  opcode[n]=o;
  opcode[m]=x;
}

long
data_push(GEN x)
{
  long n=stack_new(&s_data);
  data[n] = gclone(x);
  return n-offset;
}

